@@ -74,6 +74,7 @@
    /* Put the tokens into the symbol table, so that GDB and other debuggers
       know about them.  */
    enum yytokentype {
+     END = 0,
      NUM = 258,
      NEWLINE = 259,
      DL = 260,
@@ -91,6 +92,7 @@
    };
 #endif
 /* Tokens.  */
+#define END 0
 #define NUM 258
 #define NEWLINE 259
 #define DL 260
@@ -113,52 +115,47 @@
 
 
 
-/* 
+/*
    IGraph library.
    Copyright (C) 2009-2012  Gabor Csardi <csardi.gabor@gmail.com>
    334 Harvard st, Cambridge, MA, 02138 USA
-   
+
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
-   
+
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
-   
+
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc.,  51 Franklin Street, Fifth Floor, Boston, MA 
+   Foundation, Inc.,  51 Franklin Street, Fifth Floor, Boston, MA
    02110-1301 USA
 
 */
 
-#include "config.h"
-
-#include "core/math.h"
 #include "internal/hacks.h"
 #include "io/dl-header.h"
 #include "io/parsers/dl-parser.h"
 #include "io/parsers/dl-lexer.h"
+#include "io/parse_utils.h"
 
-#include <stdio.h>
-
-int igraph_dl_yyerror(YYLTYPE* locp, igraph_i_dl_parsedata_t* context, 
-		      const char *s);
-int igraph_i_dl_add_str(char *newstr, int length, 
-			igraph_i_dl_parsedata_t *context);
-int igraph_i_dl_add_edge(long int from, long int to,
-			 igraph_i_dl_parsedata_t *context);
-int igraph_i_dl_add_edge_w(long int from, long int to, 
-			   igraph_real_t weight,
-			   igraph_i_dl_parsedata_t *context);
-
-extern igraph_real_t igraph_pajek_get_number(const char *str, long int len);
+int igraph_dl_yyerror(YYLTYPE* locp, igraph_i_dl_parsedata_t* context,
+                      const char *s);
+static igraph_error_t igraph_i_dl_add_str(char *newstr, yy_size_t length,
+                        igraph_i_dl_parsedata_t *context);
+static igraph_error_t igraph_i_dl_add_edge(igraph_integer_t from, igraph_integer_t to,
+                         igraph_i_dl_parsedata_t *context);
+static igraph_error_t igraph_i_dl_add_edge_w(igraph_integer_t from, igraph_integer_t to,
+                           igraph_real_t weight,
+                           igraph_i_dl_parsedata_t *context);
+static igraph_error_t igraph_i_dl_check_vid(igraph_integer_t dl_vid);
 
 #define scanner context->scanner
- 
+
 
 
 /* Enabling traces.  */
@@ -183,7 +180,7 @@ extern igraph_real_t igraph_pajek_get_number(const char *str, long int len);
 typedef union YYSTYPE
 
 {
-  long int integer;
+  igraph_integer_t integer;
   igraph_real_t real;
 }
 /* Line 193 of yacc.c.  */
@@ -526,13 +523,13 @@ static const yytype_int8 yyrhs[] =
 /* YYRLINE[YYN] -- source line where rule number YYN was defined.  */
 static const yytype_uint16 yyrline[] =
 {
-       0,   108,   108,   110,   110,   112,   112,   114,   115,   116,
-     119,   119,   121,   121,   123,   124,   125,   128,   129,   135,
-     135,   140,   140,   142,   152,   154,   156,   156,   158,   162,
-     166,   171,   175,   177,   178,   179,   180,   181,   184,   185,
-     188,   190,   194,   197,   198,   201,   203,   207,   210,   226,
-     228,   229,   230,   231,   232,   235,   236,   239,   241,   244,
-     244,   250,   251,   254,   256,   260,   260
+       0,   104,   104,   114,   114,   116,   116,   118,   119,   120,
+     123,   123,   125,   125,   127,   128,   129,   132,   133,   139,
+     139,   144,   144,   146,   161,   163,   165,   165,   167,   171,
+     175,   180,   184,   186,   187,   188,   189,   190,   193,   194,
+     197,   202,   209,   217,   218,   221,   223,   227,   235,   254,
+     256,   257,   258,   259,   260,   263,   264,   267,   269,   276,
+     276,   284,   285,   288,   290,   294,   294
 };
 #endif
 
@@ -541,18 +538,20 @@ static const yytype_uint16 yyrline[] =
    First, the terminals, then, starting at YYNTOKENS, nonterminals.  */
 static const char *const yytname[] =
 {
-  "$end", "error", "$undefined", "NUM", "NEWLINE", "DL", "NEQ", "DATA",
-  "LABELS", "LABELSEMBEDDED", "FORMATFULLMATRIX", "FORMATEDGELIST1",
-  "FORMATNODELIST1", "DIGIT", "LABEL", "EOFF", "ERROR", "$accept", "input",
-  "trail", "eof", "rest", "formfullmatrix", "newline", "fullmatrix",
-  "labels", "fullmatrixdata", "zerooneseq", "zeroone",
-  "labeledfullmatrixdata", "reallabeledfullmatrixdata", "labelseq",
-  "label", "labeledmatrixlines", "labeledmatrixline", "edgelist1",
-  "edgelist1rest", "edgelist1data", "edgelist1dataline", "integer",
-  "labelededgelist1data", "labelededgelist1dataline", "weight", "elabel",
-  "nodelist1", "nodelist1rest", "nodelist1data", "nodelist1dataline",
-  "from", "tolist", "labelednodelist1data", "labelednodelist1dataline",
-  "fromelabel", "labeltolist", 0
+  "\"end of file\"", "error", "$undefined", "\"number\"",
+  "\"end of line\"", "\"DL\"", "\"n=vertexcount\"", "\"data:\"",
+  "\"labels:\"", "\"labels embedded:\"", "FORMATFULLMATRIX",
+  "FORMATEDGELIST1", "FORMATNODELIST1", "\"binary digit\"", "\"label\"",
+  "EOFF", "ERROR", "$accept", "input", "trail", "eof", "rest",
+  "formfullmatrix", "newline", "fullmatrix", "labels", "fullmatrixdata",
+  "zerooneseq", "zeroone", "labeledfullmatrixdata",
+  "reallabeledfullmatrixdata", "labelseq", "label", "labeledmatrixlines",
+  "labeledmatrixline", "edgelist1", "edgelist1rest", "edgelist1data",
+  "edgelist1dataline", "integer", "labelededgelist1data",
+  "labelededgelist1dataline", "weight", "elabel", "nodelist1",
+  "nodelist1rest", "nodelist1data", "nodelist1dataline", "from", "tolist",
+  "labelednodelist1data", "labelednodelist1dataline", "fromelabel",
+  "labeltolist", 0
 };
 #endif
 
@@ -1548,7 +1547,15 @@ yyreduce:
     {
         case 2:
 
-    { context->n=(yyvsp[(3) - (7)].integer); ;}
+    {
+  context->n=(yyvsp[(3) - (7)].integer);
+  if (context->n < 0) {
+    IGRAPH_YY_ERRORF("Invalid vertex count in DL file (%" IGRAPH_PRId ").", IGRAPH_EINVAL, context->n);
+  }
+  if (context->n > IGRAPH_DL_MAX_VERTEX_COUNT) {
+    IGRAPH_YY_ERRORF("Vertex count too large in DL file (%" IGRAPH_PRId ").", IGRAPH_EINVAL, context->n);
+  }
+;}
     break;
 
   case 7:
@@ -1598,10 +1605,10 @@ yyreduce:
 
   case 18:
 
-    { 
-	      igraph_i_dl_add_str(igraph_dl_yyget_text(scanner), 
-                                  igraph_dl_yyget_leng(scanner), 
-				  context); ;}
+    {
+              IGRAPH_YY_CHECK(igraph_i_dl_add_str(igraph_dl_yyget_text(scanner),
+                                                  igraph_dl_yyget_leng(scanner),
+                                                  context)); ;}
     break;
 
   case 19:
@@ -1625,11 +1632,16 @@ yyreduce:
   case 23:
 
     {
-  if (igraph_dl_yyget_text(scanner)[0]=='1') {
-    IGRAPH_CHECK(igraph_vector_push_back(&context->edges, 
-					 context->from));
-    IGRAPH_CHECK(igraph_vector_push_back(&context->edges, 
-					 context->to));
+  /* TODO: What if the digit is neither 0 or 1? Are multigraphs allowed? */
+  char c = igraph_dl_yyget_text(scanner)[0];
+  if (c == '1') {
+    IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges,
+                                         context->from));
+    IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges,
+                                         context->to));
+  } else if (c != '0') {
+      IGRAPH_YY_ERRORF("Unexpected digit '%c' in adjacency matrix in DL file.",
+                    IGRAPH_EINVAL, c);
   }
   context->to += 1;
 ;}
@@ -1647,24 +1659,24 @@ yyreduce:
 
   case 28:
 
-    { igraph_i_dl_add_str(igraph_dl_yyget_text(scanner), 
-                                   igraph_dl_yyget_leng(scanner), 
-				   context); ;}
+    { IGRAPH_YY_CHECK(igraph_i_dl_add_str(igraph_dl_yyget_text(scanner),
+                                                   igraph_dl_yyget_leng(scanner),
+                                                   context)); ;}
     break;
 
   case 29:
 
     {
-	         context->from += 1; 
-		 context->to = 0;
+                 context->from += 1;
+                 context->to = 0;
                ;}
     break;
 
   case 30:
 
-    { 
-	         context->from += 1; 
-		 context->to = 0;
+    {
+                 context->from += 1;
+                 context->to = 0;
                ;}
     break;
 
@@ -1716,20 +1728,31 @@ yyreduce:
   case 40:
 
     {
-                   igraph_i_dl_add_edge_w((yyvsp[(1) - (4)].integer)-1, (yyvsp[(2) - (4)].integer)-1, (yyvsp[(3) - (4)].real), context); ;}
+                    igraph_integer_t from = (yyvsp[(1) - (4)].integer), to = (yyvsp[(2) - (4)].integer);
+                    IGRAPH_YY_CHECK(igraph_i_dl_check_vid(from));
+                    IGRAPH_YY_CHECK(igraph_i_dl_check_vid(to));
+                    IGRAPH_YY_CHECK(igraph_i_dl_add_edge_w(from-1, to-1, (yyvsp[(3) - (4)].real), context)); ;}
     break;
 
   case 41:
 
     {
-		   igraph_i_dl_add_edge((yyvsp[(1) - (3)].integer)-1, (yyvsp[(2) - (3)].integer)-1, context);
+                    igraph_integer_t from = (yyvsp[(1) - (3)].integer), to = (yyvsp[(2) - (3)].integer);
+                    IGRAPH_YY_CHECK(igraph_i_dl_check_vid(from));
+                    IGRAPH_YY_CHECK(igraph_i_dl_check_vid(to));
+                    IGRAPH_YY_CHECK(igraph_i_dl_add_edge(from-1, to-1, context));
 ;}
     break;
 
   case 42:
 
-    { (yyval.integer)=igraph_pajek_get_number(igraph_dl_yyget_text(scanner), 
-					  igraph_dl_yyget_leng(scanner)); ;}
+    {
+    igraph_integer_t val;
+    IGRAPH_YY_CHECK(igraph_i_parse_integer(igraph_dl_yyget_text(scanner),
+                                           igraph_dl_yyget_leng(scanner),
+                                           &val));
+    (yyval.integer)=val;
+;}
     break;
 
   case 43:
@@ -1745,36 +1768,44 @@ yyreduce:
   case 45:
 
     {
-                          igraph_i_dl_add_edge_w((yyvsp[(1) - (4)].integer), (yyvsp[(2) - (4)].integer), (yyvsp[(3) - (4)].real), context); ;}
+                          IGRAPH_YY_CHECK(igraph_i_dl_add_edge_w((yyvsp[(1) - (4)].integer), (yyvsp[(2) - (4)].integer), (yyvsp[(3) - (4)].real), context)); ;}
     break;
 
   case 46:
 
     {
-			  igraph_i_dl_add_edge((yyvsp[(1) - (3)].integer), (yyvsp[(2) - (3)].integer), context);
+                          IGRAPH_YY_CHECK(igraph_i_dl_add_edge((yyvsp[(1) - (3)].integer), (yyvsp[(2) - (3)].integer), context));
  ;}
     break;
 
   case 47:
 
-    { (yyval.real)=igraph_pajek_get_number(igraph_dl_yyget_text(scanner), 
-					 igraph_dl_yyget_leng(scanner)); ;}
+    {
+    igraph_real_t val;
+    IGRAPH_YY_CHECK(igraph_i_parse_real(igraph_dl_yyget_text(scanner),
+                                        igraph_dl_yyget_leng(scanner),
+                                        &val));
+    (yyval.real)=val;
+;}
     break;
 
   case 48:
 
     {
+  igraph_integer_t trie_id;
+
   /* Copy label list to trie, if needed */
   if (igraph_strvector_size(&context->labels) != 0) {
-    long int i, id, n=igraph_strvector_size(&context->labels);
+    igraph_integer_t i, id, n=igraph_strvector_size(&context->labels);
     for (i=0; i<n; i++) {
-      igraph_trie_get(&context->trie,
-		      STR(context->labels, i), &id);
+      IGRAPH_YY_CHECK(igraph_trie_get(&context->trie, igraph_strvector_get(&context->labels, i), &id));
     }
     igraph_strvector_clear(&context->labels);
   }
-  igraph_trie_get2(&context->trie, igraph_dl_yyget_text(scanner), 
-		   igraph_dl_yyget_leng(scanner), &(yyval.integer));
+  IGRAPH_YY_CHECK(igraph_trie_get_len(&context->trie, igraph_dl_yyget_text(scanner),
+                                   igraph_dl_yyget_leng(scanner), &trie_id));
+  IGRAPH_ASSERT(0 <= trie_id && trie_id < IGRAPH_DL_MAX_VERTEX_COUNT);
+  (yyval.integer) = trie_id;
  ;}
     break;
 
@@ -1825,8 +1856,12 @@ yyreduce:
 
   case 58:
 
-    { context->from=igraph_pajek_get_number(igraph_dl_yyget_text(scanner),
-							  igraph_dl_yyget_leng(scanner)); ;}
+    {
+  IGRAPH_YY_CHECK(igraph_i_parse_integer(igraph_dl_yyget_text(scanner),
+                  igraph_dl_yyget_leng(scanner),
+                  &context->from));
+  IGRAPH_YY_CHECK(igraph_i_dl_check_vid(context->from));
+;}
     break;
 
   case 59:
@@ -1836,10 +1871,12 @@ yyreduce:
 
   case 60:
 
-    { 
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, 
-				       context->from-1)); 
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, (yyvsp[(2) - (2)].integer)-1));
+    {
+  igraph_integer_t to = (yyvsp[(2) - (2)].integer);
+  IGRAPH_YY_CHECK(igraph_i_dl_check_vid(to));
+  IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges,
+                                              context->from-1));
+  IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges, to-1));
  ;}
     break;
 
@@ -1868,9 +1905,9 @@ yyreduce:
   case 66:
 
     {
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, 
-				       context->from));
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, (yyvsp[(2) - (2)].integer)));
+  IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges,
+                                              context->from));
+  IGRAPH_YY_CHECK(igraph_vector_int_push_back(&context->edges, (yyvsp[(2) - (2)].integer)));
  ;}
     break;
 
@@ -2099,43 +2136,57 @@ yyreturn:
 
 
 
-int igraph_dl_yyerror(YYLTYPE* locp, igraph_i_dl_parsedata_t* context, 
-		      const char *s) {
-  snprintf(context->errmsg, 
-	   sizeof(context->errmsg)/sizeof(char)-1, 
-	   "%s in line %i", s, locp->first_line);
+int igraph_dl_yyerror(YYLTYPE* locp,
+                      igraph_i_dl_parsedata_t* context,
+                      const char *s) {
+  snprintf(context->errmsg, sizeof(context->errmsg)/sizeof(char)-1,
+           "Parse error in DL file, line %i (%s)",
+           locp->first_line, s);
   return 0;
 }
 
-int igraph_i_dl_add_str(char *newstr, int length, 
-			igraph_i_dl_parsedata_t *context) {
-  int tmp=newstr[length];
-  newstr[length]='\0';
-  IGRAPH_CHECK(igraph_strvector_add(&context->labels, newstr));
-  newstr[length]=tmp;
-  return 0;
+static igraph_error_t igraph_i_dl_add_str(char *newstr, yy_size_t length,
+                        igraph_i_dl_parsedata_t *context) {
+  IGRAPH_CHECK(igraph_strvector_push_back_len(&context->labels, newstr, length));
+  return IGRAPH_SUCCESS;
 }
 
-int igraph_i_dl_add_edge(long int from, long int to, 
-			 igraph_i_dl_parsedata_t *context) {
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, from));
-  IGRAPH_CHECK(igraph_vector_push_back(&context->edges, to));
-  return 0;
+static igraph_error_t igraph_i_dl_add_edge(igraph_integer_t from, igraph_integer_t to,
+                         igraph_i_dl_parsedata_t *context) {
+  //IGRAPH_CHECK(igraph_i_dl_check_vid(from+1));
+  //IGRAPH_CHECK(igraph_i_dl_check_vid(to+1));
+  IGRAPH_CHECK(igraph_vector_int_push_back(&context->edges, from));
+  IGRAPH_CHECK(igraph_vector_int_push_back(&context->edges, to));
+  return IGRAPH_SUCCESS;
 }
 
-int igraph_i_dl_add_edge_w(long int from, long int to, 
-			   igraph_real_t weight,
-			   igraph_i_dl_parsedata_t *context) {
-  long int n=igraph_vector_size(&context->weights);
-  long int n2=igraph_vector_size(&context->edges)/2;
+static igraph_error_t igraph_i_dl_add_edge_w(igraph_integer_t from, igraph_integer_t to,
+                           igraph_real_t weight,
+                           igraph_i_dl_parsedata_t *context) {
+  igraph_integer_t n=igraph_vector_size(&context->weights);
+  igraph_integer_t n2=igraph_vector_int_size(&context->edges)/2;
   if (n != n2) {
-    igraph_vector_resize(&context->weights, n2);
+    IGRAPH_CHECK(igraph_vector_resize(&context->weights, n2));
     for (; n<n2; n++) {
       VECTOR(context->weights)[n]=IGRAPH_NAN;
     }
   }
   IGRAPH_CHECK(igraph_i_dl_add_edge(from, to, context));
   IGRAPH_CHECK(igraph_vector_push_back(&context->weights, weight));
-  return 0;
+  return IGRAPH_SUCCESS;
+}
+
+/* Raise an error if the vertex index is invalid in the DL file.
+ * DL files use 1-based vertex indices. */
+static igraph_error_t igraph_i_dl_check_vid(igraph_integer_t dl_vid) {
+    if (dl_vid < 1) {
+        IGRAPH_ERRORF("Invalid vertex index in DL file: %" IGRAPH_PRId ".",
+                      IGRAPH_EINVAL, dl_vid);
+    }
+    if (dl_vid > IGRAPH_DL_MAX_VERTEX_COUNT) {
+        IGRAPH_ERRORF("Vertex index too large in DL file: %" IGRAPH_PRId ".",
+                      IGRAPH_EINVAL, dl_vid);
+    }
+    return IGRAPH_SUCCESS;
 }
 
